#include <stdio.h>
#include "graph.h"
#include "pnm.h"
#include "yuv.h"
#include <math.h>
#include <string.h>


#define CYCLE 3     //Frequency of Graph Cuts

#define min(a, b) (((a) < (b)) ? (a) : (b))
#define max(a, b) (((a) > (b)) ? (a) : (b))



void main(int argc, char *argv[])
{
	int i, j, d, source;
	int left_pixel_x, center_pixel_x, right_pixel_x;
	int error_l, error_r;
	int k;
	int mode; //0-->disparity mode, 1-->depth mode

	FILE *cam_param;
	FILE *input_param;

	char cfg_path[80], *ppath = cfg_path;

	char name_cam_param[64];
	char name_input_param[64];

	char *p;

	cparam cleft;
	cparam ccenter;
	cparam cright;

	Image *dis;
	char name_dis[64];

	int MIN_D = 0;
	int MAX_D = 0;
	int MAX_RANGE = 0; //upper bound on disparity range
	int MIN_RANGE = 0; //lower bound on disparity range
	double lambda; //Smoothing Coefficient (real value >=1.0)


/*------------- Input Parameter Files -------------*/

	int w;
	int h;
	int total_frame;
	int n;

	char s11[48];
	char s12[48];
	char s21[48];
	char s22[48];
	char s31[48];
	char s32[48];

	s11[0] = 0;
	s12[0] = 0;
	s21[0] = 0;
	s22[0] = 0;
	s31[0] = 0;
	s32[0] = 0;

	char ivname_l[48];
	char ivname_c[48];
	char ivname_r[48];

	ivname_l[0] = 0;
	ivname_c[0] = 0;
	ivname_r[0] = 0;


	strcpy(cfg_path, argv[0]);
	ppath = strrchr(cfg_path, '.');
	ppath++;
	strcpy(&*ppath, "cfg");

	if((input_param = fopen(cfg_path, "r")) == NULL)
		printf("can not open the param file\n");
	else {
		fscanf(input_param, "%d", &mode);
		fscanf(input_param, "%d%d", &w, &h);
		fscanf(input_param, "%d", &total_frame);
		fscanf(input_param, "%s%s%s", s11, s21, s31);
		fscanf(input_param, "%d%d", &MIN_D, &MAX_D); 
		fscanf(input_param, "%d%d", &MIN_RANGE, &MAX_RANGE);
		fscanf(input_param, "%lf", &lambda);
		fscanf(input_param, "%s%s%s", ivname_l, ivname_c, ivname_r);
		fscanf(input_param, "%s", name_dis);
		fscanf(input_param, "%s", name_cam_param);
	}

	if(mode == 0)
	printf("Compute depth from camera\n");
	else
	printf("Compute depth from the origin of 3D space\n");

	putchar('\n');

	const int WIDTH = w;
	const int HEIGHT = h;
	printf("Image width = %d\n", WIDTH);
	printf("Image height = %d\n", HEIGHT);

	putchar('\n');

	printf("tota number of frames = %d\n", total_frame);

	putchar('\n');

	putchar('\n');

	if((cam_param = fopen(name_cam_param, "r")) == NULL)
		printf("can not open the param file\n");
	else {

		fscanf(cam_param, "%s", s12);
	
		while((p=strstr(s11, s12)) == NULL){
			fscanf(cam_param, "%s", s12);
		}

		printf("Reading Camera Parameters\n");

		printf("%s \n", s12);

		fscanf(cam_param, "%lf%lf%lf", &cleft.intrinsic[0][0], &cleft.intrinsic[0][1], &cleft.intrinsic[0][2]);
		fscanf(cam_param, "%lf%lf%lf", &cleft.intrinsic[1][0], &cleft.intrinsic[1][1], &cleft.intrinsic[1][2]);
		fscanf(cam_param, "%lf%lf%lf", &cleft.intrinsic[2][0], &cleft.intrinsic[2][1], &cleft.intrinsic[2][2]);
		fscanf(cam_param, "%lf%lf", &cleft.xy[0][0], &cleft.xy[1][0]);
		fscanf(cam_param, "%lf%lf%lf%lf", &cleft.extrinsic[0][0], &cleft.extrinsic[0][1], &cleft.extrinsic[0][2], &cleft.extrinsic[0][3]);
		fscanf(cam_param, "%lf%lf%lf%lf", &cleft.extrinsic[1][0], &cleft.extrinsic[1][1], &cleft.extrinsic[1][2], &cleft.extrinsic[1][3]);
		fscanf(cam_param, "%lf%lf%lf%lf", &cleft.extrinsic[2][0], &cleft.extrinsic[2][1], &cleft.extrinsic[2][2], &cleft.extrinsic[2][3]);

		printf("Left camera parameter: \n");
		printf("%lf %lf %lf \n", cleft.intrinsic[0][0], cleft.intrinsic[0][1], cleft.intrinsic[0][2]);
		printf("%lf %lf %lf \n", cleft.intrinsic[1][0], cleft.intrinsic[1][1], cleft.intrinsic[1][2]);
		printf("%lf %lf %lf \n", cleft.intrinsic[2][0], cleft.intrinsic[2][1], cleft.intrinsic[2][2]);
		printf("%lf \n %lf \n", cleft.xy[0][0], cleft.xy[1][0]);
		printf("%lf %lf %lf %lf \n", cleft.extrinsic[0][0], cleft.extrinsic[0][1], cleft.extrinsic[0][2], cleft.extrinsic[0][3]);
		printf("%lf %lf %lf %lf \n", cleft.extrinsic[1][0], cleft.extrinsic[1][1], cleft.extrinsic[1][2], cleft.extrinsic[1][3]);
		printf("%lf %lf %lf %lf \n", cleft.extrinsic[2][0], cleft.extrinsic[2][1], cleft.extrinsic[2][2], cleft.extrinsic[2][3]);

		putchar('\n');

////////////////////////////////////////

		fscanf(cam_param, "%s\n", s22);

		while((p=strstr(s21, s22)) == NULL){
			fscanf(cam_param, "%s", s22);
		}

		printf("%s \n", s22);

		fscanf(cam_param, "%lf%lf%lf", &ccenter.intrinsic[0][0], &ccenter.intrinsic[0][1], &ccenter.intrinsic[0][2]);
		fscanf(cam_param, "%lf%lf%lf", &ccenter.intrinsic[1][0], &ccenter.intrinsic[1][1], &ccenter.intrinsic[1][2]);
		fscanf(cam_param, "%lf%lf%lf", &ccenter.intrinsic[2][0], &ccenter.intrinsic[2][1], &ccenter.intrinsic[2][2]);
		fscanf(cam_param, "%lf%lf", &ccenter.xy[0][0], &ccenter.xy[1][0]);
		fscanf(cam_param, "%lf%lf%lf%lf", &ccenter.extrinsic[0][0], &ccenter.extrinsic[0][1], &ccenter.extrinsic[0][2], &ccenter.extrinsic[0][3]);
		fscanf(cam_param, "%lf%lf%lf%lf", &ccenter.extrinsic[1][0], &ccenter.extrinsic[1][1], &ccenter.extrinsic[1][2], &ccenter.extrinsic[1][3]);
		fscanf(cam_param, "%lf%lf%lf%lf", &ccenter.extrinsic[2][0], &ccenter.extrinsic[2][1], &ccenter.extrinsic[2][2], &ccenter.extrinsic[2][3]);

		printf("Center camera parameter: \n");
		printf("%lf %lf %lf \n", ccenter.intrinsic[0][0], ccenter.intrinsic[0][1], ccenter.intrinsic[0][2]);
		printf("%lf %lf %lf \n", ccenter.intrinsic[1][0], ccenter.intrinsic[1][1], ccenter.intrinsic[1][2]);
		printf("%lf %lf %lf \n", ccenter.intrinsic[2][0], ccenter.intrinsic[2][1], ccenter.intrinsic[2][2]);
		printf("%lf \n %lf \n", ccenter.xy[0][0], ccenter.xy[1][0]);
		printf("%lf %lf %lf %lf \n", ccenter.extrinsic[0][0], ccenter.extrinsic[0][1], ccenter.extrinsic[0][2], ccenter.extrinsic[0][3]);
		printf("%lf %lf %lf %lf \n", ccenter.extrinsic[1][0], ccenter.extrinsic[1][1], ccenter.extrinsic[1][2], ccenter.extrinsic[1][3]);
		printf("%lf %lf %lf %lf \n", ccenter.extrinsic[2][0], ccenter.extrinsic[2][1], ccenter.extrinsic[2][2], ccenter.extrinsic[2][3]);

		putchar('\n');

////////////////////////////////////////

		fscanf(cam_param, "%s", s32);

		while((p=strstr(s31, s32)) == NULL){
			fscanf(cam_param, "%s", s32);
		}

		printf("%s \n", s32);

		fscanf(cam_param, "%lf%lf%lf", &cright.intrinsic[0][0], &cright.intrinsic[0][1], &cright.intrinsic[0][2]);
		fscanf(cam_param, "%lf%lf%lf", &cright.intrinsic[1][0], &cright.intrinsic[1][1], &cright.intrinsic[1][2]);
		fscanf(cam_param, "%lf%lf%lf", &cright.intrinsic[2][0], &cright.intrinsic[2][1], &cright.intrinsic[2][2]);
		fscanf(cam_param, "%lf%lf", &cright.xy[0][0], &cright.xy[1][0]);
		fscanf(cam_param, "%lf%lf%lf%lf", &cright.extrinsic[0][0], &cright.extrinsic[0][1], &cright.extrinsic[0][2], &cright.extrinsic[0][3]);
		fscanf(cam_param, "%lf%lf%lf%lf", &cright.extrinsic[1][0], &cright.extrinsic[1][1], &cright.extrinsic[1][2], &cright.extrinsic[1][3]);
		fscanf(cam_param, "%lf%lf%lf%lf", &cright.extrinsic[2][0], &cright.extrinsic[2][1], &cright.extrinsic[2][2], &cright.extrinsic[2][3]);

		printf("Right camera parameter: \n");
		printf("%lf %lf %lf \n", cright.intrinsic[0][0], cright.intrinsic[0][1], cright.intrinsic[0][2]);
		printf("%lf %lf %lf \n", cright.intrinsic[1][0], cright.intrinsic[1][1], cright.intrinsic[1][2]);
		printf("%lf %lf %lf \n", cright.intrinsic[2][0], cright.intrinsic[2][1], cright.intrinsic[2][2]);
		printf("%lf \n %lf \n", cright.xy[0][0], cright.xy[1][0]);
		printf("%lf %lf %lf %lf \n", cright.extrinsic[0][0], cright.extrinsic[0][1], cright.extrinsic[0][2], cright.extrinsic[0][3]);
		printf("%lf %lf %lf %lf \n", cright.extrinsic[1][0], cright.extrinsic[1][1], cright.extrinsic[1][2], cright.extrinsic[1][3]);
		printf("%lf %lf %lf %lf \n", cright.extrinsic[2][0], cright.extrinsic[2][1], cright.extrinsic[2][2], cright.extrinsic[2][3]);
	}

	putchar('\n');


	printf("Minimum value of disparity search range (integer): ");
	printf("%d \n", MIN_D);

	printf("Maximum value of disparity search range (integer): ");
	printf("%d \n", MAX_D);

	putchar('\n');

	printf("Minimum value of disparity range (integer): ");
	printf("%d \n", MIN_RANGE);

	printf("Maximum value of disparity range (integer): ");
	printf("%d \n", MAX_RANGE);

	putchar('\n');

	printf("Smoothing Coefficient (real number >= 1.0): ");
	printf("%lf \n", lambda);

	putchar('\n');


	Image *im_left_yuv;
	Image *im_center_yuv;
	Image *im_right_yuv;

	char name_left[128];
	char name_center[128];
	char name_right[128];

	const int LABEL_NUMBER = MAX_D - min(MIN_D, 0) +1;

	double ratio;

	if(abs(ccenter.extrinsic[0][3] - cleft.extrinsic[0][3]) >= abs(ccenter.extrinsic[0][3] - cright.extrinsic[0][3]))
		ratio = abs(ccenter.extrinsic[0][3] - cleft.extrinsic[0][3]) / abs(ccenter.extrinsic[0][3] - cright.extrinsic[0][3]);
	else
		ratio = abs(ccenter.extrinsic[0][3] - cright.extrinsic[0][3]) / abs(ccenter.extrinsic[0][3] - cleft.extrinsic[0][3]);

	int cycle;

	printf("Depth map of center image (YUV 4:2:0 format): %s", name_dis);

	putchar('\n');
	putchar('\n');

	sprintf(name_left,  "%s", ivname_l); 
	sprintf(name_center,"%s", ivname_c);
	sprintf(name_right, "%s", ivname_r);

    im_left_yuv = new_image(WIDTH, HEIGHT, 3);
    im_center_yuv = new_image(WIDTH, HEIGHT, 3);
    im_right_yuv = new_image(WIDTH, HEIGHT, 3);

	dis = new_image(WIDTH, HEIGHT, 1);

	FILE *fpi_l, *fpi_c, *fpi_r;
	FILE *fpo;

	long position = 0L;

	if((fpi_l=fopen(name_left, "rb"))==NULL || (fpi_c=fopen(name_center, "rb"))==NULL || (fpi_r=fopen(name_right, "rb"))==NULL || (fpo=fopen(name_dis, "wb"))==NULL){
		fprintf(stderr, "Can't open YUV image files.\n");
		exit(1);
	} else {
		for(n=0; n<total_frame; n++){

/*------------- Read YUV 4:2:0 Image -------------*/

			printf("Reading YUV 4:2:0 Image \n");
			printf("frame number = %d \n", n);

			if(fseek(fpi_l, position, SEEK_SET)){
				exit(1);
			} else {
				fread(im_left_yuv->data, sizeof(unsigned char), 3 * WIDTH * HEIGHT / 2, fpi_l);
			}

			if(fseek(fpi_c, position, SEEK_SET)){
				exit(1);
			} else {
				fread(im_center_yuv->data, sizeof(unsigned char), 3 * WIDTH * HEIGHT / 2, fpi_c);
			}

			if(fseek(fpi_r, position, SEEK_SET)){
				exit(1);
			} else {
				fread(im_right_yuv->data, sizeof(unsigned char), 3 * WIDTH * HEIGHT/ 2, fpi_r);
			}

			position = ftell(fpi_l); 

			putchar('\n');


/*------------- Luminance -------------*/

			cycle = CYCLE;

			int *FLAG;
			int *label;
			int *ERROR;
	
			Graph::node_id *nodes;
			Graph::node_id *auxiliary;

			int *t_link_a, *t_link_b;


			const int NODE_NUMBER = HEIGHT * WIDTH;

			nodes     = (Graph::node_id *)malloc(NODE_NUMBER* sizeof(Graph::node_id));
			auxiliary = (Graph::node_id *)malloc((NODE_NUMBER<<1)*sizeof(Graph::node_id));

			t_link_a = (int *)calloc(NODE_NUMBER, sizeof(int));
			t_link_b = (int *)calloc(NODE_NUMBER, sizeof(int));

			ERROR = (int *)calloc(NODE_NUMBER * LABEL_NUMBER, sizeof(int));
			if(ERROR == NULL) {
				puts("can not secure memory space");
			}

			FLAG = (int *)calloc(NODE_NUMBER, sizeof(int));
			if(FLAG == NULL)
				puts("can not secure memory space");

			label = (int *)calloc(NODE_NUMBER, sizeof(int));
			if(label == NULL)
				puts("can not secure memory space");


/*------- Block Matching -------*/

			printf("Block Matching\n");
	
			for(j = 0; j < HEIGHT; j++){
				for(i = 0; i < WIDTH; i++){
					for(d = MIN_D; d <= MAX_D; d++){
						error_l = 0;
						error_r = 0;
						center_pixel_x = i;

						if(abs(ccenter.extrinsic[0][3] - cleft.extrinsic[0][3]) >= abs(ccenter.extrinsic[0][3] - cright.extrinsic[0][3])){
							left_pixel_x   = i + (int)(d * ratio + 0.5);
							right_pixel_x  = i - d;
						} else {
							left_pixel_x   = i + d;
							right_pixel_x  = i - (int)(d * ratio + 0.5);
						}


						if(left_pixel_x < 0 || left_pixel_x >= WIDTH){
							error_l = 255;	
						}else{
							error_l = abs(im_left_yuv->data[j * WIDTH + left_pixel_x] - im_center_yuv->data[j * WIDTH + center_pixel_x]);
						}

						if(right_pixel_x < 0 || right_pixel_x >= WIDTH){
							error_r = 255;	
						}else{
							error_r = abs(im_center_yuv->data[j * WIDTH + center_pixel_x] - im_right_yuv->data[j * WIDTH + right_pixel_x]);
						}

						ERROR[(j * WIDTH + i)*LABEL_NUMBER + d - min(MIN_D, 0)] = min(error_l, error_r);
					}//i
				}//j
			}//d

/*------------- Graph Cuts -------------*/
			printf("Graph Cuts\n");
			for(j = 0; j < HEIGHT; j++){
				for(i = 0; i < WIDTH; i++){
					label[j * WIDTH + i] = MIN_D;
				}
			}

//			printf("Cycle count down \n");

			while(cycle != 0){
//				printf("%d \n", cycle); //count down cycle

				for(source = MIN_D; source <= MAX_D; source++){ //integer "source" moves from MIN_D to MAX_D.
					Graph *g = new Graph();
					int counter = 0;

					for(j = 0; j < HEIGHT; j++){
						for(i = 0; i < WIDTH; i++){
							nodes[j * WIDTH + i] = g -> add_node();
							//if L=(alpha)CFLAG=1
							if(label[j * WIDTH + i] == source) FLAG[j * WIDTH + i] = 1;
							else	FLAG[j * WIDTH + i] = 0;
						}
					}

					for(j = 0; j < HEIGHT; j++){
						for(i = 0; i < WIDTH; i++){

							// t_link of NODE
							t_link_a[j * WIDTH + i] = ERROR[(j * WIDTH + i)*LABEL_NUMBER + source - min(MIN_D, 0)];
							if(FLAG[j * WIDTH + i] == 0)	t_link_b[j * WIDTH + i] = ERROR[(j * WIDTH + i)*LABEL_NUMBER + label[j * WIDTH + i] - min(MIN_D, 0)];
							else							t_link_b[j * WIDTH + i] = 1<<10;
							g -> set_tweights(nodes[j * WIDTH + i], t_link_a[j * WIDTH + i], t_link_b[j * WIDTH + i]);


							// add auxiliary node
							if(j == HEIGHT-1 && i == WIDTH-1){
								;
							}else if(j == HEIGHT-1){
								if(label[j * WIDTH + i] != label[j * WIDTH + i + 1]){
									auxiliary[counter] = g -> add_node();
									g -> set_tweights(auxiliary[counter], 0, (int)(lambda * abs(label[j * WIDTH + i] - label[j * WIDTH + i + 1])));
									g -> add_edge(nodes[j * WIDTH + i], auxiliary[counter], (int)(lambda * abs(label[j * WIDTH + i] - source)), (int)(lambda * abs(label[j * WIDTH + i] - source)));
									g -> add_edge(auxiliary[counter], nodes[j * WIDTH + i + 1], (int)(lambda * abs(label[j * WIDTH + i + 1] - source)), (int)(lambda * abs(label[j * WIDTH + i + 1] - source)));
									counter ++;
								}else{
									g -> add_edge(nodes[j * WIDTH + i], nodes[j * WIDTH + i + 1], (int)(lambda * abs(label[j * WIDTH + i] - source)), (int)(lambda * abs(label[j * WIDTH + i] - source)));
								}
							}else if(i == WIDTH-1){
								if(label[j * WIDTH + i] != label[(j + 1) * WIDTH + i]){
									auxiliary[counter] = g -> add_node();
									g -> set_tweights(auxiliary[counter], 0, (int)(lambda * abs(label[j * WIDTH + i] - label[(j + 1) * WIDTH + i])));
									g -> add_edge(nodes[j * WIDTH + i], auxiliary[counter], (int)(lambda * abs(label[j * WIDTH + i] - source)), (int)(lambda * abs(label[j * WIDTH + i] - source)));
									g -> add_edge(auxiliary[counter], nodes[(j + 1) * WIDTH + i], (int)(lambda * abs(label[(j + 1) * WIDTH + i] - source)), (int)(lambda * abs(label[(j + 1) * WIDTH + i] - source)));
									counter ++;
								}else{
									g -> add_edge(nodes[j * WIDTH + i], nodes[(j + 1) * WIDTH + i], (int)(lambda * abs(label[j * WIDTH + i] - source)), (int)(lambda * abs(label[j * WIDTH + i] - source)));
								}
							}else{
								if(label[j * WIDTH + i] != label[j * WIDTH + i + 1]){
									auxiliary[counter] = g -> add_node();
									g -> set_tweights(auxiliary[counter], 0, (int)(lambda * abs(label[j * WIDTH + i] - label[j * WIDTH + i + 1])));
									g -> add_edge(nodes[j * WIDTH + i], auxiliary[counter], (int)(lambda * abs(label[j * WIDTH + i] - source)), (int)(lambda * abs(label[j * WIDTH + i] - source)));
									g -> add_edge(auxiliary[counter], nodes[j * WIDTH + i + 1], (int)(lambda * abs(label[j * WIDTH + i + 1] - source)), (int)(lambda * abs(label[j * WIDTH + i + 1] - source)));
									counter ++;
								}else{
									g -> add_edge(nodes[j * WIDTH + i], nodes[j * WIDTH + i + 1], (int)(lambda * abs(label[j * WIDTH + i] - source)), (int)(lambda * abs(label[j * WIDTH + i] - source)));
								}
								if(label[j * WIDTH + i] != label[(j + 1) * WIDTH + i]){
									auxiliary[counter] = g -> add_node();
									g -> set_tweights(auxiliary[counter], 0, (int)(lambda * abs(label[j * WIDTH + i] - label[(j + 1) * WIDTH + i])));
									g -> add_edge(nodes[j * WIDTH + i], auxiliary[counter], (int)(lambda * abs(label[j * WIDTH + i] - source)), (int)(lambda * abs(label[j * WIDTH + i] - source)));
									g -> add_edge(auxiliary[counter], nodes[(j + 1) * WIDTH + i], (int)(lambda * abs(label[(j + 1) * WIDTH+i] - source)), (int)(lambda * abs(label[(j + 1) * WIDTH + i] - source)));
									counter ++;
								}else{
									g -> add_edge(nodes[j * WIDTH + i], nodes[(j + 1) * WIDTH + i], (int)(lambda * abs(label[j * WIDTH + i] - source)), (int)(lambda * abs(label[j * WIDTH + i] - source)));
								}
							}
						}
					}

					printf("cycle:%d, label=(%d), auxiliary:%d\n", CYCLE - cycle + 1, source, counter);

					Graph::flowtype flow = g -> maxflow();

					for(j = 0; j < HEIGHT; j++){
						for(i = 0; i < WIDTH; i++){
							if (g->what_segment(nodes[j * WIDTH + i]) == Graph::SOURCE){
							}else{
								label[j * WIDTH + i] = source;
							}
						}
					}

					delete g;
				}

				for(j = 0; j < HEIGHT; j++){
					for(i = 0; i < WIDTH; i++){
						dis->data[j * WIDTH + i] = label[j * WIDTH + i];
					}
				}


				if(cycle == 1){
					if(mode == 0){ //disparity mode
						if(MIN_D >= 1){
							for(j = 0; j < HEIGHT; j++){
								for(i = 0; i < WIDTH; i++){
									dis->data[j * WIDTH + i] = (int)(255*(1.0 * (dis->data[j * WIDTH + i]) - MIN_RANGE) / ( 1.0 * (MAX_RANGE - MIN_RANGE))+0.5);
								}
							}
						} else {
							for(j = 0; j < HEIGHT; j++){
								for(i = 0; i < WIDTH; i++){
									dis->data[j * WIDTH + i] = dis->data[j * WIDTH + i] - MIN_D + 1;
								}
							}
							for(j = 0; j < HEIGHT; j++){
								for(i = 0; i < WIDTH; i++){
									dis->data[j * WIDTH + i] = (int)(255.0 * (1.0 * dis->data[j * WIDTH + i] - 1.0) / (1.0 * (MAX_RANGE - MIN_RANGE))+0.5);
								}
							}
						} 
					} //disparity mode
					if(mode == 1){ //depth mode
						if(MIN_D >= 1){
							for(j = 0; j < HEIGHT; j++){
								for(i = 0; i < WIDTH; i++){
									dis->data[j * WIDTH + i] = (int)(255.0 * ((ccenter.extrinsic[2][3])*(MAX_RANGE +(ccenter.intrinsic[0][2] - cleft.intrinsic[0][2])) - (ccenter.intrinsic[0][0])*(ccenter.extrinsic[0][3] - cleft.extrinsic[0][3]))*(dis->data[j * WIDTH + i] - MIN_RANGE) / (((ccenter.extrinsic[2][3])*(dis->data[j * WIDTH + i] +(ccenter.intrinsic[0][2] - cleft.intrinsic[0][2])) - (ccenter.intrinsic[0][0])*(ccenter.extrinsic[0][3] - cleft.extrinsic[0][3])) * (MAX_RANGE - MIN_RANGE))+0.5);
								}
							}
						} else {
							for(j = 0; j < HEIGHT; j++){
								for(i = 0; i < WIDTH; i++){
									dis->data[j * WIDTH + i] = dis->data[j * WIDTH + i] - MIN_D + 1;
								}
							}
							for(j = 0; j < HEIGHT; j++){
								for(i = 0; i < WIDTH; i++){
									dis->data[j * WIDTH + i] = (int)(255.0 * ((ccenter.extrinsic[2][3])*(MAX_RANGE +(ccenter.intrinsic[0][2] - cleft.intrinsic[0][2]))+(ccenter.intrinsic[0][0])*(ccenter.extrinsic[0][3] - cleft.extrinsic[0][3]))*(dis->data[j * WIDTH + i] - 1.0 + (MIN_D - MIN_RANGE)) / (((ccenter.extrinsic[2][3])*(dis->data[j * WIDTH + i] - 1 + MIN_RANGE +(ccenter.intrinsic[0][2] - cleft.intrinsic[0][2]))+(ccenter.intrinsic[0][0])*(ccenter.extrinsic[0][3] - cleft.extrinsic[0][3])) * (MAX_RANGE - MIN_RANGE))+0.5);
								}
							}
						} 
					} //depth mode

/* --- write depth map into YUV file --- */

					for (int jj = 0; jj < HEIGHT; jj++) {
						for (int ii = 0; ii < WIDTH; ii++) {
							fputc((unsigned char)(dis->data[jj * WIDTH + ii]), fpo);
						}
					}

					for (int jj = 0; jj < HEIGHT/2; jj++) {
						for (int ii = 0; ii < WIDTH/2; ii++) {
							fputc((unsigned char)128, fpo);
						}
					}

					for (int jj = 0; jj < HEIGHT/2; jj++) {
						for (int ii = 0; ii < WIDTH/2; ii++) {
							fputc((unsigned char)128, fpo);
						}
					}
				} //if(cycle==1)
				cycle -= 1;
				printf("\n");
	} //while(cycle != 0)

	free(nodes);
	free(auxiliary);

	free(t_link_a);
	free(t_link_b);


	free(FLAG);
	free(label);

	free(ERROR);
		} // for n
}

	free_image(dis);
	free_image(im_left_yuv);
	free_image(im_center_yuv);
	free_image(im_right_yuv);

/* ---------------------- z values ---------------------- */

	char zv_file_name[] = "z_value.txt"; //z value output file
	FILE *zf;

	if ((zf = fopen(zv_file_name, "w")) == NULL)
		printf("can not open file");
	else{
		if(mode == 0){
			printf("Depth map from camera is outputted. \n"); //Indicate z values from camera
			printf("Max and Min of depth from camera: \n");
			printf("Z_near from camera = %lf \n", (ccenter.intrinsic[0][0])*(ccenter.extrinsic[0][3] - cleft.extrinsic[0][3])/(MAX_RANGE + ccenter.intrinsic[0][2] - cleft.intrinsic[0][2]));
			printf("Z_far  from camera = %lf \n", (ccenter.intrinsic[0][0])*(ccenter.extrinsic[0][3] - cleft.extrinsic[0][3])/(MIN_RANGE + ccenter.intrinsic[0][2] - cleft.intrinsic[0][2]));
			printf("Input these values to View Synthesis software.\n");

			fprintf(zf, "Z_near from camera = "); //Output z values to a text file
			fprintf(zf, "%lf\n", (ccenter.intrinsic[0][0])*(ccenter.extrinsic[0][3] - cleft.extrinsic[0][3])/(MAX_RANGE + ccenter.intrinsic[0][2] - cleft.intrinsic[0][2]));
			fprintf(zf, "Z_far from camera = ");
			fprintf(zf, "%lf\n", (ccenter.intrinsic[0][0])*(ccenter.extrinsic[0][3] - cleft.extrinsic[0][3])/(MIN_RANGE + ccenter.intrinsic[0][2] - cleft.intrinsic[0][2]));
			fprintf(zf, "Input these values to View Synthesis software.\n");
		}
		if(mode == 1){
			printf("Depth map from the origin of 3D space is outputted. \n"); //Indicate z values from the origin of 3D space
			printf("Max and Min of depth from origin of 3D space: \n");
			printf("Z_near from the origin of 3D space = %lf \n", (ccenter.intrinsic[2][3]) + (ccenter.intrinsic[0][0])*(ccenter.extrinsic[0][3] - cleft.extrinsic[0][3])/(MAX_RANGE + ccenter.intrinsic[0][2] - cleft.intrinsic[0][2]));
			printf("Z_far  from the origin of 3D space = %lf \n", (ccenter.intrinsic[2][3]) + (ccenter.intrinsic[0][0])*(ccenter.extrinsic[0][3] - cleft.extrinsic[0][3])/(MIN_RANGE + ccenter.intrinsic[0][2] - cleft.intrinsic[0][2]));
			printf("Input these values to View Synthesis software.\n");

			fprintf(zf, "Z_near from the origin of 3D space = "); //Out put z values to a text file
			fprintf(zf, "%lf\n", (ccenter.extrinsic[2][3]) + (ccenter.intrinsic[0][0])*(ccenter.extrinsic[0][3] - cleft.extrinsic[0][3])/(MAX_RANGE + ccenter.intrinsic[0][2] - cleft.intrinsic[0][2]));
			fprintf(zf, "Z_far from the origin of 3D space = ");
			fprintf(zf, "%lf\n", (ccenter.extrinsic[2][3]) + (ccenter.intrinsic[0][0])*(ccenter.extrinsic[0][3] - cleft.extrinsic[0][3])/(MIN_RANGE + ccenter.intrinsic[0][2] - cleft.intrinsic[0][2]));
			fprintf(zf, "Input these values to View Synthesis software.\n");
		}
	}

	putchar('\n');

	fclose(fpi_l);
	fclose(fpi_c);
	fclose(fpi_r);
	fclose(fpo);
	fclose(cam_param);
	fclose(input_param);
	fclose(zf);
}
